use std::fmt;

use mitra_models::users::types::User;
use mitra_utils::{
    crypto_eddsa::{
        ed25519_public_key_from_private_key,
        Ed25519PrivateKey,
    },
    did_key::DidKey,
};

fn fep_ef61_identity(secret_key: &Ed25519PrivateKey) -> String {
    let public_key = ed25519_public_key_from_private_key(secret_key);
    let did_key = DidKey::from_ed25519_key(public_key.as_bytes());
    format!(
        "did:ap:key:{}",
        did_key.key_multibase(),
    )
}

pub enum Authority {
    Server(String),
    ServerKey((String, Ed25519PrivateKey)),
}

impl fmt::Display for Authority {
    fn fmt(&self, formatter: &mut fmt::Formatter<'_>) -> fmt::Result {
        let authority_str = match self {
            Self::Server(ref server_url) => server_url.to_owned(),
            Self::ServerKey((_, secret_key)) => {
                fep_ef61_identity(secret_key)
            },
        };
        write!(formatter, "{}", authority_str)
    }
}

impl Authority {
    pub fn server(server_url: &str) -> Self {
        Self::Server(server_url.to_owned())
    }

    pub fn server_key(server_url: &str, secret_key: &Ed25519PrivateKey) -> Self {
        Self::ServerKey((server_url.to_owned(), *secret_key))
    }

    pub fn from_user(
        server_url: &str,
        user: &User,
        fep_ef61_enabled: bool,
    ) -> Self {
        if fep_ef61_enabled {
            Self::server_key(server_url, &user.ed25519_private_key)
        } else {
            Self::server(server_url)
        }
    }

    pub fn is_fep_ef61(&self) -> bool {
        !matches!(self, Self::Server(_))
    }

    pub fn server_url(&self) -> &str {
        match self {
            Self::Server(ref server_url) => server_url,
            Self::ServerKey((ref server_url, _)) => server_url,
        }
    }
}

#[cfg(test)]
mod tests {
    use mitra_utils::{
        crypto_eddsa::generate_weak_ed25519_key,
        urls::Url,
    };
    use super::*;

    const SERVER_URL: &str = "https://server.example";

    #[test]
    fn test_authority() {
        let authority = Authority::server(SERVER_URL);
        assert!(!authority.is_fep_ef61());
        assert_eq!(authority.to_string(), "https://server.example");
        assert_eq!(authority.server_url(), SERVER_URL);
    }

    #[test]
    fn test_authority_fep_ef61() {
        let secret_key = generate_weak_ed25519_key();
        let authority = Authority::server_key(SERVER_URL, &secret_key);
        assert!(authority.is_fep_ef61());
        assert_eq!(authority.to_string(), "did:ap:key:z6MkvUie7gDQugJmyDQQPhMCCBfKJo7aGvzQYF2BqvFvdwx6");
        assert_eq!(authority.server_url(), SERVER_URL);
    }

    #[test]
    fn test_fep_ef61_identity_url_compat() {
        let secret_key = generate_weak_ed25519_key();
        let did_ap_key = fep_ef61_identity(&secret_key);
        let did_url = format!("{did_ap_key}/objects/1");
        assert_eq!(did_url, "did:ap:key:z6MkvUie7gDQugJmyDQQPhMCCBfKJo7aGvzQYF2BqvFvdwx6/objects/1");
        let url = Url::parse(&did_url).unwrap();
        assert_eq!(url.scheme(), "did");
        assert_eq!(url.authority(), "");
        assert_eq!(url.path(), "ap:key:z6MkvUie7gDQugJmyDQQPhMCCBfKJo7aGvzQYF2BqvFvdwx6/objects/1");
        assert_eq!(url.to_string(), did_url);
    }
}
